;-------------------------------------------------------------------------------
;
;  CRT.inc
;  programmed by Marco Kaufmann
;  Release 2002/09/16
;
;  Include file for CodeX Assembler programs,
;  requires Mem.inc
;
;  CRT.inc provides support of basic crt routines, such as cursor control,
;  keyboard input, screen output and sound
;
;  CRT.inc is written for 16bit DOS applications, but does also work in
;  32bit PM DOS applications, as well as probably in Win32 COM applications
;  (or also not).
;
;  This file as part of the CodeX Assembler package is distributed free,
;  but without any warranty, even without the warranty of merchantability.
;  I remind, that use is on own risk.
;
;-------------------------------------------------------------------------------


;---mode_xxx Constants for textmodes--------------------------------------------


  mode_40x25 equ 1
  mode_80x25 equ 3
  mode_40x50 equ 101
  mode_80x50 equ 103


;---cursor_xxx constants for cursor sizes---------------------------------------


  cursor_none     equ 0FFFFh
  cursor_normal   equ 90Ah
  cursor_half     equ 50Ah
  cursor_upnormal equ 1
  cursor_uphalf   equ 5      ;does not work on some systems, shows full cursor instead of
  cursor_full     equ 0Ah


;---col_xxx constants for VGA colors--------------------------------------------


  col_black      equ 0
  col_blue       equ 1
  col_green      equ 2
  col_cyan       equ 3
  col_red        equ 4
  col_lime       equ 5
  col_brown      equ 6
  col_grey       equ 7
  col_darkgrey   equ 8
  col_lightblue  equ 9
  col_lightgreen equ 0Ah
  col_lightcyan  equ 0Bh
  col_lightred   equ 0Ch
  col_lightlime  equ 0Dh
  col_yellow     equ 0Eh
  col_white      equ 0Fh



;---Pseudo interface------------------------------------------------------------
;   These are the routines provided by crt.inc,
;   whereby routines marked by "*" are implemented yet:
;
;
;     initcrt
;
;   * setvideomode
;   * getvideomode
;
;   * textcolor
;   * textbackground
;   * clrscr
;     clrscrattr
;     clreol
;     clreolattr
;
;   * printchar
;   * print
;     println
;     putchar
;     put
;     putline
;     flushmask
;
;     scolup
;     scrolldown
;
;   * setcursorsize
;   * gotoxy
;   * inccursorpos
;   * cursorpos
;
;   * readkey
;   * keypressed
;     getline
;
;   * sound
;   * nosound
;   * delay
;
;
;
;   These are the variables provided by crt.inc, but also used internal:


  textattr db 7         ;current text attributes, fg=grey, bk=black by default

  textw    db 80        ;current screen width, 80 by default
  texth    db 25        ;current screen height, 25 by default

  vidseg   dw 0B800h    ;current segment of video buffer (0B800h by default, DOS)

  tmp1     resd 1
  tmp2     resd 1
  tmp3     resd 1

  screensizeshl db 1    ;value shifting 40x25 left with to get the size of
                        ;the screenbuffer (1 by default, for 80x25)


;---SetVideoMode----------------------------------------------------------------
;   param : ax = mode (mode_xxx)
;   return: none


setvideomode:

  cmp al,1
  jz @msetvideomode2
  mov screensizeshl,1
  mov textw,80
  jmp @msetvideomode3

  @msetvideomode2:
  mov screensizeshl,0
  mov textw,40
  @msetvideomode3:
  mov texth,25

  mov word tmp1, ax
  xor ah,ah
  int 10h

  mov ax,word tmp1
  test ax,100h
  jz @msetvideomode1
  add screensizeshl,1
  add texth,25
  mov ax,1112h
  xor bl,bl
  int 10h

  @msetvideomode1:
  ret


;---GetVideoMode----------------------------------------------------------------
;   param : none
;   return: ax = mode (mode_xxx)


getvideomode:

  mov bx,40h
  mov gs,bx
  mov bl,gs:[84h]

  cmp bl,0
  jnz @mgetvideomode1
  mov bl,25

  @mgetvideomode1:
  mov bh,gs:[4Ah]

  mov ax,1
  cmp bh,80
  jnz @mgetvideomode2
  mov al,3

  @mgetvideomode2:
  cmp bl,25
  jbe @mgetvideomode3
  mov ah,1

  @mgetvideomode3:
  ret


;---TextColor-------------------------------------------------------------------
;   param : al = col (col_xxx)
;   return: none


textcolor:

  and textattr,0F0h
  or textattr,al
  ret


;---TextBackGround--------------------------------------------------------------
;   param : al = col (col_xxx)
;   return: none


textbackground:

  shl al,4
  and textattr,0Fh
  or textattr,al
  ret


;---SetCursorSize---------------------------------------------------------------
;   param : none
;   return: cx = size (cursor_xxx or userdefined)


setcursorsize:

  mov ax,100h
  int 10h
  ret


;---GotToXY---------------------------------------------------------------------
;   param : dl = x, dh = y
;   return: none


gotoxy:

  mov ax,200h
  xor bx,bx
  int 10h
  ret


;---CursorPos-------------------------------------------------------------------
;   param : none
;   return: al = x, ah = y


cursorpos:

  mov ax,40h
  mov gs,ax

  mov al,gs:[50h]
  mov ah,gs:[51h]
  ret


;---IncCursorPos----------------------------------------------------------------
;   param : none
;   return: none


inccursorpos:

  call cursorpos
  inc al
  mov dx,ax
  call gotoxy

  call cursorpos
  cmp al,texth
  ret


;---ReadKey---------------------------------------------------------------------
;   param : none
;   return: al = ASCII char


readkey:

  mov ax,1000h
  int 16h

  cmp al,0E0h
  jnz @mreadkey1
  cmp ah,0
  jz @mreadkey1
  xor al,ah

  @mreadkey1:
  ret


;---KeyPressed------------------------------------------------------------------
;   param : none
;   return: al = boolean


keypressed:

  mov ax,1100h
  int 16h

  jz @mkeypressed1
  mov al,1
  jmp @mkeypressed2
  @mkeypressed1:
  xor al,al
  @mkeypressed2:
  ret


;---Sound-----------------------------------------------------------------------
;   param : ecx = freq
;   return: none


sound:

  mov al,0B6h
  out 43h,al       ;put the PIT to the frequency in ecx
  xor edx,edx
  mov eax,1234DDh
  div ecx          ;devide freq by PIT freq (1234DDh)
  out 42h,al
  shr ax,8
  out 42h,al

  in al,61h        ;set bits 0,1 of PC speaker control register:
  or al,3          ;turn on and connect to PIT
  out 61h,al
  ret


;---NoSound---------------------------------------------------------------------
;   param : none
;   return: none


nosound:

  in al,61h        ;turn off the control flags:
  and al,0FCh      ;turn of speaker and disconnect from PIT (not
  out 61h,al       ;neccessary to reprogram the PIT)
  ret


;---Delay-----------------------------------------------------------------------
;   param : ecx = time
;   return: none


delay:

  jecxz @mdelay1
  @mdelay2:
  mov tmp1,ecx

  mov ecx,10000h
  @mdelay3:
  dec ecx
  jnz @mdelay3

  mov ecx,tmp1
  dec ecx
  jnz @mdelay2
  @mdelay1:
  ret


;---PrintChar-------------------------------------------------------------------
;   param : al = ASCII char
;   return: none


printchar:

  mov ah,9
  xor bh,bh
  mov bl,textattr
  mov cx,1
  int 10h
  ret


;---Print-----------------------------------------------------------------------
;   param : edx = offset string
;   return: none


print:

  xor ecx,ecx
  mov cl,[edx]

  jecxz @mprintstring2
  xor bx,bx
  mov bl,textattr

  @mprintstring1:
  mov word tmp2,bx
  mov tmp1,ecx

  mov ah,9
  inc edx
  mov al,[edx]
  mov tmp3,edx
  mov cx,1
  int 10h
  call inccursorpos

  mov edx,tmp3
  mov ecx,tmp1
  mov bx,word tmp2
  dec ecx
  jnz @mprintstring1

  @mprintstring2:
  ret


;---ClrScr----------------------------------------------------------------------
;   param : none
;   return: none


clrscr:
  mov eax,40*25*2
  mov cl,screensizeshl
  shl eax,cl
  mov ecx,eax

  xor eax,eax           ;fill screenbuffer with ' ' and textattr
  mov gs,vidseg
  mov bh,textattr
  mov bl,' '
  call fillwordfar

  xor dx,dx
  jmp gotoxy            ;set cursor to 0,0 (can use jmp instead of call, ret)
  ret
