;-------------------------------------------------------------------------------
;
;  VGA.inc
;  programmed by Marco Kaufmann
;  Release 2002/09/29
;
;  Include file for CodeX Assembler programs
;
;  VGA.inc provides few VGA graphical routines, adapted from
;  WarpGraph (programmed by Marco Kaufmann) for FreePascal
;
;  This file as part of the CodeX Assembler package is distributed free,
;  but without any warranty, even without the warranty of merchantability.
;  I remind, that use is on own risk.
;
;-------------------------------------------------------------------------------

;---Pseudo interface------------------------------------------------------------
;   These are the routines provided by vga.inc
;
;
;   setcolor
;   setbkcolor
;   setvgamode
;   hline08b
;   vline08b


;---Graphical mode - constants--------------------------------------------------

vgamode_320x200x8 equ 0;
vgamode_640x480x4 equ 1;
vgamode_640x350x4 equ 2;
vgamode_640x200x4 equ 3;
vgamode_320x200x4 equ 4;


;---VGAMode - structure, VGA modes----------------------------------------------

struc tvgamode
  width word
  height word
  bitsperpix byte
  modeid word
endstruc

vgamodes define tvgamode (320,200,8,13h),(640,480,4,12h),(640,350,4,10h),(640,200,4,0Eh),(320,200,4,0Dh)


;---Internal data---------------------------------------------------------------

screenmaxx resw 1
screenmaxy resw 1
vgascreenw resw 1
vgascreenh resw 1
vgabppix   resb 1
vgashl     resb 1 ;160 shl vgashl = linesize in bytes

vgacolor   resb 1
vgabkcolor resb 1

actscreenbank resw 1
x resd 1
y resd 1
x2 resd 1
y2 resd 1


;---SetColor--------------------------------------------------------------------
;   param : cl = color
;   return: none

setcolor:
  mov vgacolor,cl
  ret


;---SetBkColor------------------------------------------------------------------
;   param : cl = color
;   return: none

setbkcolor:
  mov vgabkcolor,cl
  ret


;---SetVGAMode------------------------------------------------------------------
;   param : ecx = vgamode_xxx
;   return: none

setvgamode:
  lea edx,[ecx*8+offset vgamodes]
  sub edx,ecx
  mov ax,[edx]
  mov bx,[edx+2]
  mov cl,[edx+4]
  mov vgascreenw,ax
  mov vgascreenh,bx
  mov vgabppix,cl
  shr cl,3
  cmp ax,640
  jnz @msetvgamode1
  inc cl
  @msetvgamode1:
  mov vgashl,cl
  dec ax
  dec bx
  mov screenmaxx,ax
  mov screenmaxy,bx
  mov word vgacolor,15h

  mov ax,[edx+5]
  int 10h

  xor dx,dx
  mov actscreenbank,dx
  mov ax,4F05h
  xor bx,bx
  int 10h                   ;set act screenbank zero
  ret


;---HLine08B--------------------------------------------------------------------
;   param : eax = x1, ebx = y, ecx = x2
;   return: none

hline08b: 

  mov x,eax
  mov y,ebx
  mov x2,ecx
  cmp bx,screenmaxy
  ja @hline08bmend
  test ax,8000h
  jz @hline08bm1
    test cx,8000h
    jnz @hline08bmend
    xor ax,ax
    mov x,ax
    cmp cx,screenmaxx
    jna @hline08bm2
    mov cx,screenmaxx
    mov x2,cx
    jmp @hline08bm2
  @hline08bm1:
    test cx,8000h
    jz @hline08bm3
    xor cx,cx
    mov x,cx
    cmp ax,screenmaxx
    jna @hline08bm2
    mov ax,screenmaxx
    mov x,ax
    jmp @hline08bm2
    @hline08bm3:
    cmp ax,screenmaxx
    jna @hline08bm4
    cmp cx,screenmaxx
    jna @hline08bm5
    jmp @hline08bmend
    @hline08bm4:
    cmp cx,screenmaxx
    jna @hline08bm2
    mov cx,screenmaxx
    mov x2,cx
    jmp @hline08bm2
    @hline08bm5:
    mov ax,screenmaxx
    mov x,ax
  @hline08bm2:
  cmp ax,cx
  jna @hline08bm6
    xchg ax,cx
    mov x,ax
  @hline08bm6:
  add cx,1
  mov x2,cx

  mov ax,vgascreenw
  mul bx         

  shl edx,16
  mov dx,ax
  mov ax,x
  and eax,0FFFFh
  add edx,eax
  xor eax,eax
  mov cx,dx
  shr edx,16
  cmp dx,actscreenbank
  jz @hline08bm21
    mov actscreenbank,dx
    mov ax,4F05h
    xor bx,bx
    int 10h
  @hline08bm21:
  mov ax,x2
  sub ax,x
  add ax,cx
  jnc @hline08bm22

  xor edx,edx
  mov dx,cx
  mov ecx,65536
  sub ecx,edx
  push gs
  push ax
  mov ax,0A000h
  mov gs,ax
  mov bl,vgacolor
  mov bh,bl
  shl ebx,8
  mov bl,bh
  shl ebx,8
  mov bl,bh

  test ecx,1
  jz @hline08bm28
    mov gs:[edx],bl
    add edx,1
  @hline08bm28:
  test ecx,2
  jz @hline08bm29
    mov gs:[edx],bx
    add edx,2
  @hline08bm29:

  shr ecx,2
  jz @hline08bmxxx
  sub ecx,1
  jz @hline08bm212
  @hline08bm211:
    mov gs:[edx+ecx*4],ebx
    loop @hline08bm211
  @hline08bm212:
  mov gs:[edx+ecx*4],ebx

  @hline08bmxxx:
  push ebx
  add actscreenbank,1
  mov dx,actscreenbank
  mov ax,4F05h
  xor bx,bx
  int 10h
  pop ebx
  xor edx,edx
  xor ecx,ecx
  pop cx
  jmp @hline08bm210

  @hline08bm22:

  sub ax,cx
  xor edx,edx
  mov dx,cx
  xor ecx,ecx
  mov cx,ax
  push gs
  mov ax,0A000h
  mov gs,ax
  mov bl,vgacolor
  mov bh,bl
  shl ebx,8
  mov bl,bh
  shl ebx,8
  mov bl,bh

  @hline08bm210:
  test ecx,1
  jz @hline08bm26
    mov gs:[edx],bl
    add edx,1
  @hline08bm26:
  test ecx,2
  jz @hline08bm27
    mov gs:[edx],bx
    add edx,2
  @hline08bm27:

  shr ecx,2
  jz @hline08bm215
  sub ecx,1
  jz @hline08bm24
  @hline08bm23:
    mov gs:[edx+ecx*4],ebx
    loop @hline08bm23
  @hline08bm24:
  mov gs:[edx+ecx],ebx
  @hline08bm215:
  pop gs

  @hline08bmend:
  ret


;---VLine08B--------------------------------------------------------------------
;   param : eax = x, ebx = y1, ecx = y2
;   return: none

vline08b:

  mov x,eax
  mov y,ebx
  mov y2,ecx
  cmp ax,screenmaxx
  ja @vline08bmend
  test bx,8000h
  jz @vline08bm1
    test cx,8000h
    jnz @vline08bmend
    xor bx,bx
    mov y,bx
    cmp cx,screenmaxy
    jna @vline08bm2
    mov cx,screenmaxy
    mov y2,cx
    jmp @vline08bm2
  @vline08bm1:
    test cx,8000h
    jz @vline08bm3
    xor cx,cx
    mov y,cx
    cmp bx,screenmaxy
    jna @vline08bm2
    mov bx,screenmaxy
    mov y,bx
    jmp @vline08bm2
    @vline08bm3:
    cmp bx,screenmaxy
    jna @vline08bm4
    cmp cx,screenmaxy
    jna @vline08bm5
    jmp @vline08bmend
    @vline08bm4:
    cmp cx,screenmaxy
    jna @vline08bm2
    mov cx,screenmaxy
    mov y2,cx
    jmp @vline08bm2
    @vline08bm5:
    mov bx,screenmaxy
    mov y,bx
  @vline08bm2:
  cmp bx,cx
  jna @vline08bm6
    xchg bx,cx
    mov y,bx
  @vline08bm6:
  add cx,1
  mov y2,cx

  mov ax,vgascreenw
  mul bx             
  shl edx,16
  mov dx,ax
  mov ax,x

  and eax,0FFFFh
  add edx,eax
  xor eax,eax
  mov cx,dx
  shr edx,16
  cmp dx,actscreenbank
  jz @vline08bm21
    mov actscreenbank,dx
    mov ax,4F05h
    xor bx,bx
    int 10h
  @vline08bm21:
  mov di,cx
  and edi,0FFFFh
  mov bl,vgacolor
  mov cx,y2
  sub cx,y
  push gs
  mov ax,0A000h
  mov gs,ax
  mov dx,vgascreenw
  and ecx,0FFFFh
  sub ecx,1
  jz @vline08bm22
  @vline08bm25:
    mov gs:[edi],bl
    add di,dx
    jnc @vline08bm24
      push bx
      push dx
      mov ax,4F05h
      xor bx,bx
      add actscreenbank,1
      mov dx,actscreenbank
      int 10h
      pop dx
      pop bx
    @vline08bm24:
    loop @vline08bm25
  @vline08bm22:
    mov gs:[edi],bl
  @vline08bm23:
  pop gs

  @vline08bmend:
  ret



