;-------------------------------------------------------------------------------
;
;  CRT32.inc
;  programmed by Marco Kaufmann
;  Release 2002/11/19
;
;  Include file for CodeX Assembler programs,
;  requires SYSTEM32.inc
;
;  CRT.inc provides support of basic crt routines, such as cursor control,
;  keyboard input, screen output and sound
;
;  CRT.inc is written for 32bit DOS applications, but does also work with
;  Win32 COM applications (or also not).
;
;  This file as part of the CodeX Assembler package is distributed free,
;  but without any warranty, even without the warranty of merchantability.
;  I remind, that use is on own risk.
;
;-------------------------------------------------------------------------------


;---mode_xxx constants for textmodes--------------------------------------------


  mode_40x25 equ 1
  mode_80x25 equ 3
  mode_40x50 equ 101
  mode_80x50 equ 103


;---cursor_xxx constants for cursor sizes---------------------------------------


  cursor_none     equ 0FFFFh
  cursor_normal   equ 90Ah
  cursor_half     equ 50Ah
  cursor_upnormal equ 1
  cursor_uphalf   equ 5      ;does not work on some systems, shows full cursor instead of
  cursor_full     equ 0Ah


;---col_xxx constants for VGA colors--------------------------------------------


  col_black      equ 0
  col_blue       equ 1
  col_green      equ 2
  col_cyan       equ 3
  col_red        equ 4
  col_lime       equ 5
  col_brown      equ 6
  col_grey       equ 7
  col_darkgrey   equ 8
  col_lightblue  equ 9
  col_lightgreen equ 0Ah
  col_lightcyan  equ 0Bh
  col_lightred   equ 0Ch
  col_lightlime  equ 0Dh
  col_yellow     equ 0Eh
  col_white      equ 0Fh


;---Pseudo interface------------------------------------------------------------
;   These are the routines provided by crt32.inc,
;   whereby routines marked by "*" are implemented yet:
;
;

struc tlstring
  buffsize  dword
  charcount dword
  chars     ptr
endstruc

echo_normal   equ 1
echo_none     equ 2
echo_password equ 3

echo_digits    equ 004h
echo_upchars   equ 008h
echo_downchars equ 010h
echo_space     equ 020h
echo_umlautsz  equ 040h
echo_mathchars equ 080h
echo_special   equ 100h
echo_all       equ 1FCh


crt_80chars    equ 01h
crt_font8x8    equ 02h
crt_blink      equ 04h
crt_newln13    equ 08h
crt_inccursor  equ 10h
crt_incvirtual equ 20h
crt_cursoron   equ 40h

crtmaxpages equ 8

struc tcursor
  x word
  y word
  cursoroffs dword
endstruc

struc twindow
           left   dword
           top    dword
           right  dword
           bottom dword

           winw    dword
           winh    dword
           winmaxx dword
           winmaxy dword

           winlinesize  dword
           winsize      dword
           winstartoffs dword
           winendoffs   dword
         endstruc

struc tcrt
  crtflags  byte
  actmode   word
  
  screenw    dword
  screenh    dword
  screenmaxx dword
  screenmaxy dword

  screenlinesize  dword
  screensize      dword
  screenstartoffs dword
  screenendoffs   dword

  window twindow

  screenpages word
  activepage  word
  visualpage  word

  textcolor      word
  textbackground word
  textattr       word

  cursorstyle word

  cursorpos tcursor dup crtmaxpages-1

  virtualpos tcursor dup crtmaxpages-1
endstruc

crt define tcrt (0,0,0,0,0,0,0,0,0,0,(0,0,0,0,0,0,0,0,0,0,0,0),0,0,0,0,0,0,0,((0,0,0),(0,0,0),(0,0,0),(0,0,0),(0,0,0),(0,0,0),(0,0,0)),((0,0,0),(0,0,0),(0,0,0),(0,0,0),(0,0,0),(0,0,0),(0,0,0)))

; * initcrt
;   getcrtflags
;   setcrtflags
;   setcrtmode
;   getcrtmode
;   getscreenw
;   getscreenh
;   getscreenmaxx
;   getscreenmaxy
;   getscreenlinesize
;   getscreensize
;   getscreenstartoffs
;   getscreenendoffs

; * setscreenwindow
;   getscreenwindow
;   getscreenwinw
;   getscreenwinh
;   getscreenwinmaxx
;   getscreenwinmaxy
;   getscreenwinlinesize
;   getscreenwinsize
;   getscreenwinstartoffs
;   getscreenwinendoffs

;   getscreenpages
;   getscreenactivepage
;   setscreenactivepage
;   getscreenvisualpage
;   setscreenvisualpage

; * textcolor
; * textbackground
; * textattr
; * gettextcolor
; * gettextbackground
; * gettextattr

;   getcrt80chars
;   setcrt80chars
;   getcrtfont8x8
;   setcrtfont8x8
;   getcrtblink
;   setcrtblink
;   getcrtln13h
;   setcrtln13h
;   getcrtinccursor
;   setcrtinccursor
;   getcrtincvirtual
;   setcrtincvirtual

; * clrscr
; * clrscrattr
; * clrscrchr
;   clreol
;   clreolattr
;   clreolchr
;   virtualclreol
;   virtualclreolattr
;   virtualclreolchr

;   printchar
; * print
;   printz
;   printl
;   println
;   printlnz
;   printlnl
;   printmask
;   flushmaskcursor

;   putchar
;   put
;   putz
;   putl
;   putln
;   putlnz
;   putlnl
;   putmask
;   flushmask

; * scrollup
; * screenscrollup

;   getcursoron
;   setcursoron
;   cursoron
;   cursoroff
;   setcursorstyle
;   getcursorstyle

; * getcursorpos
; * gotoxy
; * gotoxyrel

; * getvirtualpos
; * virtualgotoxy
; * virtualgotoxyrel

; * readkey               ;wait for keystroke and read key 
; * keypressed            ;true, if key is pressed

;   getstring(dword offs, dword maxlength)   ; if maxlength is 0, a buffer is reserved on the heap and returned in offs
;   getstringz(dword offs, dword maxlength)  ; ...
;   getstringl(dword offs, dword maxlength)
;   getstringext(dword offs, dword maxlength, byte echolength, byte echochar) ; ..., echolength = size of echowindow, echochar = 0: normal echo, 0FFh: no echo, else: echochar is echo for every got char (password)
;   getstringzext(dword offs, dword maxlength, byte echolength, byte echochar)
;   getstringlext(dword offs, dword maxlength, byte echolength, byte echochar)

; * sound(dword freq)
; * nosound
;   delay(dword msec)
;   delayreal(real sec)
;   getdelaybase
;   setdelaybase

;   crthline(word x1, word y, word x2, word transparent)
;   crtvline(word x1, word y1, word y2, word transparent)
;   crtrect(word x1, word y1, word x2, word y2, word transparent)
;   doublecrthline
;   doublecrtvline
;   doublecrtrect

;   strbin
;   strdec
;   strhex
;   valbin
;   valdec
;   valhex
;   ascii2asciil
;   ascii2asciiz
;   asciil2ascii
;   asciil2asciiz
;   asciiz2ascii
;   asciiz2asciil


proc initcrt
  mov ax,1130h                          ;get count of crt lines
  xor bx,bx
  int 10h
  and edx,0FFh
  mov crt.screenmaxy,edx
  inc edx
  mov crt.screenh,edx

  mov gs,defsegs.biosmemseg             ;get current video mode
  mov bl,gs:[4Ah]
  and ebx,0FFh
  mov crt.screenw,ebx
  mov ecx,ebx
  shl cx,1
  mov crt.screenlinesize,cx
  dec ebx
  mov crt.screenmaxx,ebx

  cmp bx,79                             ;set screenpages, screensize and screenflags according to crt solution
  jnz @m40cols
  cmp edx,25
  ja @m80x50
  mov word crt.screensize,4000
  mov word crt.screenpages,8
  mov word crt.crtflags,crt_80chars |& 38h
  mov word crt.actmode,mode_80x25
  jmp @m2
  @m80x50:
  mov word crt.screensize,8000
  mov word crt.screenpages,4
  mov word crt.crtflags,crt_80chars |& crt_font8x8 |& 38h
  mov word crt.actmode,mode_80x50
  jmp @m2
  @m40cols:
  cmp edx,25
  ja @m40x50
  mov word crt.screensize,2000
  mov word crt.screenpages,8
  mov word crt.crtflags,38h
  mov word crt.actmode,mode_40x25
  jmp @m2
  @m40x50:
  mov word crt.screensize,4000
  mov word crt.screenpages,8
  mov word crt.crtflags,crt_font8x8 |& 38h
  mov word crt.actmode,mode_40x50
  @m2:

  mov al,gs:[62h]                       ;set screenstartoffs and screenendoffs
  mov crt.visualpage,al
  mov crt.activepage,al
  and ax,0FFh
  imul ax,crt.screensize
  mov crt.screenstartoffs,ax
  add ax,crt.screensize
  mov crt.screenendoffs,ax 

  push offset crt.virtualpos            ;reset virtual cursors
  push 8*crtmaxpages
  push word 0
  call fillchar

  mov word crt.textcolor,7              ;reset attributes
  mov word crt.textbackground,0
  mov word crt.textattr,7

  xor ebx,ebx                           ;get cursor position on all pages available
  @m3:

  mov ax,300h
  push ebx

  int 10h
  pop ebx
  mov [ebx*8+offset crt.cursorpos],dl
  mov [ebx*8+(offset crt.cursorpos+2)],dh

  mov ecx,crt.screensize
  mov al,dh
  and eax,0FFh

  imul eax,crt.screenlinesize
  shl dl,1
  and edx,0FFh
  add eax,edx

  imul ecx,ebx
  add ecx,eax
  mov [ebx*8+(offset crt.cursorpos+4)],ecx
  inc ebx
  cmp bl,8
  jnz @m3

  mov ax,gs:[60h]                       ;get cursor style
  mov crt.cursorstyle,ax
  test ah,60h
  jz @m4
  or byte crt.crtflags,crt_cursoron     ;check, if cursor is on
  @m4:

  test byte gs:[65h],20h                ;check, if blink bit is set, or if background color is intensified by bit 7 of attr byte
  jz @mnoblink
  or byte crt.crtflags,crt_blink
  @mnoblink:

  push 0
  push 0
  push dword crt.screenmaxx
  push dword crt.screenmaxy
  call setscreenwindow
  add esp,26
endproc

proc setscreenwindow(dword x1, dword y1, dword x2, dword y2)
  mov eax,x1
  mov ebx,y1
  mov ecx,x2
  mov edx,y2
  mov crt.window..left,eax
  mov crt.window..top,ebx
  mov crt.window..right,ecx
  mov crt.window..bottom,edx
  sub ecx,eax
  mov crt.window..winmaxx,ecx
  sub edx,ebx
  mov crt.window..winmaxy,edx
  inc ecx
  inc edx
  mov crt.window..winw,ecx
  mov crt.window..winh,edx

  shl ecx,1
  mov crt.window..winlinesize,ecx
  imul ecx,edx
  mov crt.window..winsize,ecx
  imul ebx,crt.screenlinesize
  shl eax,1
  add ebx,eax
  add ebx,crt.screenstartoffs
  mov crt.window..winstartoffs,ebx
  mov ecx,crt.window..winh
  imul ecx,crt.screenlinesize
  add ecx,ebx
  mov crt.window..winendoffs,ecx
endproc

proc getstringlines                     ;internal use only; offset to string in esi, length in ebx, start val of cursor x in dl, return new lines in eax,
  xor eax,eax
  xor ecx,ecx
  cmp ebx,0
  jz @mend

  @m1:
  cmp byte [esi+ecx],13
  jnz @m2

  test byte crt.crtflags,crt_newln13
  jnz @m3

  @m2:
  inc dl
  cmp dl,crt.window..winw
  jb @m4
  @m3:

  inc eax
  xor dl,dl

  @m4:
  inc ecx
  cmp ecx,ebx
  jnz @m1

  @mend:
endproc

proc textcolor(word color)
  mov al,color
  mov crt.textcolor,al
  and byte crt.textattr,0F0h
  or crt.textattr,al
endproc

proc textbackground(word color)
  mov al,color
  mov crt.textbackground,al
  shl al,4
  and byte crt.textattr,0Fh
  or crt.textattr,al
endproc

proc textattr(word attr)
  mov al,attr
  mov crt.textattr,al
  mov ah,al
  and al,0Fh
  shr ah,4
  mov crt.textcolor,al
  mov crt.textbackground,ah
endproc

proc gettextcolor                       ;return in al
  mov al,crt.textcolor
endproc

proc gettextbackground
  mov al,crt.textbackground
endproc

proc gettextattr
  mov al,crt.textbackground
endproc

proc clrscr
  mov edi,crt.window..winstartoffs
  mov ebx,crt.window..winlinesize
  mov ecx,crt.window..winh
  mov edx,crt.screenlinesize
  mov al,' '
  mov ah,crt.textattr

  @m1:
  pushad
  push word defsegs.crtseg
  push edi
  push ebx
  push ax
  call fillwordfar
  add esp,6
  pop edi
  pop ax
  popad
  add edi,edx
  dec ecx
  jnz @m1

endproc

tmpdw1 dd 0                             ;used internal

proc clrscrattr
  mov edi,crt.window..winstartoffs
  mov ebx,crt.window..winlinesize
  mov ecx,crt.window..winh
  mov edx,crt.screenlinesize
  xor eax,eax
  mov ah,crt.textattr
  shl eax,16
  mov ah,crt.textattr
  mov tmpdw1,eax

  @m1:
  pushad
  push word defsegs.crtseg
  push edi
  push ebx
  push 0FF00FFh
  call andbuffer
  pop ebx
  push tmpdw1
  call orbuffer
  add esp,8
  pop edi
  pop ax
  popad
  add edi,edx
  dec ecx
  jnz @m1

endproc

proc clrscrchar
  mov edi,crt.window..winstartoffs
  mov ebx,crt.window..winlinesize
  mov ecx,crt.window..winh
  mov edx,crt.screenlinesize
  xor eax,eax
  mov tmpdw1,200020h

  @m1:
  pushad
  push word defsegs.crtseg
  push edi
  push ebx
  push 0FF00FF00h
  call andbuffer
  pop ebx
  push tmpdw1
  call orbuffer
  add esp,8
  pop edi
  pop ax
  popad
  add edi,edx
  dec ecx
  jnz @m1

endproc

proc scrollup(word lines)
  mov cx,lines
  and ecx,0FFFFh
  mov edx,ecx
  imul edx,crt.screenlinesize
  mov eax,crt.window..winstartoffs
  add edx,eax

  cmp cx,crt.window..winh
  jb @m2

  mov cl,' '
  mov ch,crt.textattr
  xor eax,eax
  jmp @m1

  @m2:
  pushad
  push word defsegs.crtseg
  push edx
  push word defsegs.crtseg
  push eax
  push dword crt.window..winlinesize
  call movefarreverse
  add esp,16
  popad

  add eax,crt.screenlinesize
  add edx,crt.screenlinesize
  cmp edx,crt.window..winendoffs
  jnz @m2

  mov cl,' '
  mov ch,crt.textattr

  @m1:
  push word defsegs.crtseg
  push eax
  push dword crt.window..winlinesize
  push cx
  call fillwordfar
  pop cx
  pop ebx
  pop eax
  pop bx
  add eax,crt.screenlinesize
  cmp eax,crt.window..winendoffs
  jnz @m1

endproc

proc screenscrollup(word lines)
  mov cx,lines
  and ecx,0FFFFh
  mov edx,ecx
  imul edx,crt.screenlinesize
  mov eax,crt.screenstartoffs
  add edx,eax

  cmp cx,crt.window..winh
  jb @m2

  mov cl,' '
  mov ch,crt.textattr
  xor eax,eax
  jmp @m1

  @m2:
  pushad
  push word defsegs.crtseg
  push edx
  push word defsegs.crtseg
  push eax
  push dword crt.screenlinesize
  call movefarreverse
  add esp,16
  popad

  add eax,crt.screenlinesize
  add edx,crt.screenlinesize
  cmp edx,crt.screenendoffs
  jnz @m2

  mov cl,' '
  mov ch,crt.textattr

  @m1:
  push word defsegs.crtseg
  push eax
  push dword crt.screenlinesize
  push cx
  call fillwordfar
  pop cx
  pop ebx
  pop eax
  pop bx
  add eax,crt.screenlinesize
  cmp eax,crt.screenendoffs
  jnz @m1

endproc

proc getcursorpos                       ;return al x ah y
  mov cx,crt.activepage
  and ecx,0FFFFh

  mov al,[ecx*8+offset crt.cursorpos]
  mov ah,[ecx*8+(offset crt.cursorpos+2)]
endproc

proc gotoxy(word x, word y)
  mov dl,x
  mov dh,y
  mov ax,crt.activepage
  and eax,0FFh
  mov [eax*8+offset crt.cursorpos],dl
  mov [eax*8+(offset crt.cursorpos+2)],dh
  add dl,crt.window..left
  add dh,crt.window..top

  push dx
  mov bh,crt.activepage
  push bx
  mov ax,200h
  int 10h
  pop ax
  and eax,0FFh
  pop dx
  xor ebx,ebx
  mov bl,y
  mov dl,x
  and edx,0FFh
  shl dl,1
  imul ebx,crt.screenlinesize
  add ebx,edx
  add ebx,crt.window..winstartoffs
mov eax,0
  mov [eax*8+(offset crt.cursorpos+4)],ebx
endproc

proc gotoxyrel(word x, word y)
  call getcursorpos
  mov bl,ah
  and bx,0FFh
  and ax,0FFh
  add bx,y
  add ax,x
  push ax
  push bx
  call gotoxy
endproc

proc getvirtualpos                      ;return al x ah y
  mov cx,crt.activepage
  and ecx,0FFFFh

  mov al,[ecx*8+offset crt.virtualpos]
  mov ah,[ecx*8+(offset crt.virtualpos+2)]
endproc

proc virtualgotoxy(word x, word y)
  mov dl,x
  mov dh,y
  mov ax,crt.activepage
  and eax,0FFh
  mov [eax*8+offset crt.virtualpos],dl
  mov [eax*8+(offset crt.virtualpos+2)],dh
  add dl,crt.window..left
  add dh,crt.window..top

  xor ebx,ebx
  mov bl,dh
  and edx,0FFh
  shl dl,1
  imul ebx,crt.screenlinesize
  add ebx,edx
  add ebx,crt.screenstartoffs
  mov [eax*8+(offset crt.cursorpos+4)],ebx
endproc

proc virtualgotoxyrel(word x, word y)
  call getvirtualpos
  mov bl,ah
  and bx,0FFh
  and ax,0FFh
  add bx,y
  add ax,x
  push ax
  push bx
  call virtualgotoxy
endproc

proc readkey                            ;al
  mov ax,1000h
  int 16h

  cmp al,0E0h
  jnz @mend
  cmp ah,0
  jz @mend
  xor al,ah

  @mend:
endproc

proc keypressed
  mov ax,1100h
  int 16h

  jz @m1
  mov al,1
  jmp @m2
  @m1:
  xor al,al
  @m2:
endproc

proc sound(dword freq)
  mov ecx,freq
  mov al,0B6h
  out 43h,al       ;put the PIT to the frequency in ecx
  xor edx,edx
  mov eax,1234DDh
  div ecx          ;devide freq by PIT freq (1234DDh)
  out 42h,al
  shr ax,8
  out 42h,al

  in al,61h        ;set bits 0,1 of PC speaker control register:
  or al,3          ;turn on and connect to PIT
  out 61h,al
endproc

proc nosound
  in al,61h        ;turn off the control flags:
  and al,0FCh      ;turn of speaker and disconnect from PIT (not
  out 61h,al       ;neccessary to reprogram the PIT)
endproc

proc putmain(ptr src, ptr dst, dword length, word cursorpos)      ;internal use only, print,printl,printz,println,printlnl,printlnz and put,putl,putz,putln,putlnl,pulnz refer to
  mov tmpdw1,0                                                    ;return: ax = new cursor x, bx = new cursor y
  xor ebx,ebx
  xor ecx,ecx
  mov bx,crt.activepage
  xor edx,edx
  mov dx,cursorpos

  push ebx
  mov ebx,length
  mov esi,src

  call getstringlines
  shr edx,8
  add edx,eax                           ;new cursor y in ecx
  pop ebx
  cmp edx,crt.window..winmaxy
  jna @mnoscroll

  sub edx,crt.window..winmaxy           ;lines to scroll in edx
  mov tmpdw1,edx
  mov eax,crt.screenlinesize
  imul eax,edx                          ;value to decrement cursor offs, since screen scrolled
  sub dst,eax

  cmp edx,crt.window..winmaxy
  ja @mclearall
  push dx
  call scrollup
  pop dx
  jmp @mnoscroll

  @mclearall:                           ;since more lines to scroll than visible on screen, screen will be cleared compleately
  call clrscr

  @mnoscroll:                           ;scrolled screen if neccessary, put string now 

  xor ebx,ebx
  xor ecx,ecx
  mov bx,crt.activepage
  mov dx,cursorpos

  mov edi,dst
  mov ebx,length
  mov esi,src

  cmp ebx,0
  jz @m5
  mov ah,crt.textattr
  mov gs,defsegs.crtseg

  @m1:
  mov al,[esi+ecx]
  cmp al,13
  jnz @m2
  test byte crt.crtflags,crt_newln13
  jz @m2

  push ebx                              ;next line
  mov ebx,crt.window..winmaxx
  sub bl,dl
  shl ebx,1
  add edi,ebx
  pop ebx
  jmp @m3

  @m2:
  mov gs:[edi+ecx*2],ax
  inc dl
  cmp dl,crt.window..winw
  jb @m4
  @m3:

  add edi,crt.screenlinesize
  sub edi,crt.window..winlinesize
  xor dl,dl                             ;next line
  inc dh

  @m4:
  inc ecx
  cmp ecx,ebx
  jnz @m1

  xor ebx,ebx                           ;calc cursorpos
  mov bl,dl
  mov eax,ebx
  mov bl,dh
  sub ebx,tmpdw1
  jmp @mend

  @m5:
  xor ebx,ebx
  mov bl,crt.activepage
  mov ax,[ebx+offset crt.cursorpos]
  mov bx,[ebx+(offset crt.cursorpos+2)]

  @mend:
endproc

proc print(ptr asciistr)                ;string, where byte asciistr is length, and chars are at (asciistr+1)^

  mov eax,asciistr
  xor ecx,ecx
  mov cl,[eax]
  inc eax
  push eax

  xor ebx,ebx
  mov bx,crt.activepage
  mov dl,[ebx*8+offset crt.cursorpos]
  mov dh,[ebx*8+(offset crt.cursorpos+2)]
  push dword [ebx*8+(offset crt.cursorpos+4)]

  push ecx
  push dx
  call putmain
  push ax
  push bx

  call gotoxy
  add esp,18
endproc

proc printl(ptr asciistrl)              ;string, where byte asciistr^ is length, and chars are at (asciistr+1)^

  mov eax,asciistrl
  push dword [eax+ offset tlstring.chars]

  xor ebx,ebx
  mov bx,crt.activepage
  mov dl,[ebx*8+offset crt.cursorpos]
  mov dh,[ebx*8+(offset crt.cursorpos+2)]
  push dword [ebx*8+(offset crt.cursorpos+4)]

  push dword [eax+ offset tlstring.charcount]
  push dx
  call putmain

  push ax
  push bx
  call gotoxy
  add esp,18
endproc