;------------------------------------------------------------------------------------------------------------------------------
;
;  SYSTEM32.inc
;  programmed by Marco Kaufmann
;  Release 2002/11/22
;
;  Include file for CodeX Assembler programs
;
;  SYSTEM32.inc provides support for heap, memory moving and filling and
;  DPMI routines as well as access to the command line and the data structures
;  set up by the PM loader
;
;  This file as part of the CodeX Assembler package is distributed free,
;  but without any warranty, even without the warranty of merchantability.
;  I remind, that use is on own risk.
;
;------------------------------------------------------------------------------------------------------------------------------


;---Pseudo interface-----------------------------------------------------------------------------------------------------------
;   These are the routines provided by SYSTEM32.inc, where * means that a procedure is implemented already,
;   and ! that it's highly recommanded to test it
;
;
;
;   * systeminit
;   * outdword
;   * halt
;   * halterror
;
;   * fillchar
;   * fillword
;   * filldword
;   * fillcharfar
;   * fillwordfar
;   * filldwordfar
;   * andbuffer
;   * orbuffer
;   * nandbuffer
;   * norbuffer
;   * xorbuffer
;   * xnorbuffer
;   * notbuffer
;   * move
;   * movefar
;   * movereverse
;   * movefarreverse
;
;     getstacksize
;     getremainstacksize
;     resizestack
;     resetstack
;
;     getheapsize
;     resizeheap
;     resetheap
;     getmem
;     freemem
;     memavail
;     maxavail
;
;     createheap
;     destroyheap
;     setheap
;     getheap
;
;     getparamcount
;     getparam
;
;   * getmeminfo      
;     getsegmentinfo   
;     getsegment       
;     freesegment      
;     getsegmenttype   
;     setsegmenttype  
;
;   * allocdescriptor
;   * allocdescriptors
;   * allocspecdescriptor
;   * freedescriptor
;   * rmseg2pmseg
;   * pmseg2rmseg
;   * getsegmentbase
;   * setsegmentbase
;   *!getsegmentlimit
;   * setsegmentlimit
;     getsegmentdpl
;     setsegmentdpl
;     getdescriptoraccessrights
;     setdescriptoraccessrights
;     duplicatedescriptor
;     getdescriptor
;     setdescriptor
;
;     getdosmemblock
;     freedosmemblock
;     resizedosmemblock
;     allocmemblock
;     freememblock
;     resizememblock
;
;     getrmintvect
;     setrmintvect
;     getcpuexceptionvect
;     setcpuexceptionvect
;     getintvect
;     setintvect
;
;   * rmint
;     callfarrmproc
;     callfarrmprociret
;
;     allocrmcallbackaddr
;     freermcallbackaddr
;     getsaverestoreaddr
;     getrawmodeswitchaddr
;
;     getdpmiversion
;     getdpmicaps
;
;     locklinearregion
;     unlocklinearregion
;     markrmregionpagable
;     relockrmregion
;     getpagesize
;     mapphysicaladdress


;---DPMI info data structure---------------------------------------------------------------------------------------------------


struc tdpmiinfo
  dpmiversion word
  dpmienterpm dword                     ;RM address to call for enter PM
  dpmienterrm dword                     ;PM address to call for enter RM (16bit only)

  dpmidatabytes word                    ;bytes of private data required by the DPMI host
endstruc


;---DPMI memory info data structure--------------------------------------------------------------------------------------------


struc tdpmimeminfo
  memavail         dword                ;largest available free block in bytes
  maxunlockedpages dword                ;maximum unlocked page allocation in pages (4KB)
  maxlockedpages   dword                ;maximum locked page allocation in pages (4KB)
  memsize          dword                ;linear address space size in 4KB
  unlockedpages    dword                ;total count of unlocked pages 
  freepages        dword                ;total count of free pages
  physicalpages    dword                ;total count of physical pages
  memfree          dword                ;free linear address space in 4KB
  displacepages    dword                ;size of paging file / partition in 4KB
  byte dup 12                           ;Reserved, set to 0FFh always
endstruc


;---Application info data structure--------------------------------------------------------------------------------------------


struc tappinfo
  codesize     dword                    ;sizeof code
  codelinbase  dword                    ;linear base address of code and data segment
  startup      dword                    ;IP start value in CS
  heapmin      dword                    ;note to linker: heapmin+codesize should be 4KB aligned!
  heapmax      dword                    ;note to linker: meapmax+codesize should be 4KB aligned
  stackmin     dword                    ;stackmin should be 4KB aligned
  stackmax     dword                    ;stackmin should be 4KB aligned

  heapsize     dword                    ;effektive heap size 
  stacksize    dword                    ;effektive stack size
  stacklinbase dword                    ;linear base address of stack segment
endstruc


;---Default segments data structure being available after systeminit - call----------------------------------------------------


struc tdefsegs
  crtseg      word                      ;PM CRT segment (phys. 0B8000h)
  vidseg      word                      ;PM video segment (phys. 0A0000h)
  dosmemseg   word                      ;128KB memory, that can be used to exchange data with DOS
  pspseg      word                      ;program PSP (containing command line arguments)
  biosmemseg  word                      ;provide access to BIOS data area                      
  flatmemseg  word                      ;provide access to address space 0..max-1 where max is systems memory at all

  appcs word                            ;application's code
  appds word                            ;data and
  appss word                            ;stack segment

  loadercs word                         ;PM loader's code
  loaderds word                         ;data and
  loaderss word                         ;stack segment

  rmcrtseg      word                    ;same as above, but for RM use
  rmvidseg      word
  rmdosmemseg   word                      
  rmpspseg      word

  rmloadercs word
  rmloaderds word
  rmloaderss word  
endstruc


;---RM registers data structure for RM ints------------------------------------------------------------------------------------


struc trmregs
  rmedi   dword
  rmesi   dword
  rmebp   dword
          dword
  rmebx   dword
  rmedx   dword
  rmecx   dword
  rmeax   dword
  rmflags word
  rmes    word
  rmds    word
  rmfs    word
  rmgs    word
  rmip    word
  rmcs    word
  rmsp    word
  rmss    word  
endstruc


;---Errorcode constants err_xxx------------------------------------------------------------------------------------------------


err_descriptorunavailable equ 8011h
err_linmemunavailable     equ 8012h
err_physmemunavailable    equ 8013h
err_backstunavailable     equ 8014h
err_badparameter          equ 8021h
err_invalidselector       equ 8022h


;---Data provided by SYSTEM32.inc---------------------------------------------------------------------------------------------------------------------------


dpmiinfo    define tdpmiinfo    (0,0,0,0)
dpmimeminfo define tdpmimeminfo (0,0,0,0,0,0,0,0,0,(0,0,0,0,0,0,0,0,0,0,0,0))
appinfo     define tappinfo     (0,0,0,0,0,0,0,0,0,0)
defsegs     define tdefsegs     (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
rmregs      define trmregs      (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)


;---SystemInit-----------------------------------------------------------------------------------------------------------------
;   return   : none
;   operation: load dpmiinfo, appinfo and defsegs - structure from PM loader, update dpmimeminfo -, setuup rmregs and
;              extend defsegs - structure and setup default heap in application's dataseg, if required

proc systeminit
  local lddpmimeminfo equ 55 +100h
  local ldrmregs      equ 103+100h      ;offset of rmregs structure in loader DS
  local lddefsegs     equ 153+100h      ;offset of appinfo structure in loader DS

  push es                               ;copy appinfo - structure from loader's data seg
  push 0
  push ds
  push offset appinfo
  push sizeof tappinfo
  call movefar
  add esp,14

  push sizeof tappinfo                  ;copy dpmiinfo - structure from loader's data seg
  push ds
  push offset dpmiinfo
  push sizeof tdpmiinfo-2
  call movefar
  add esp,14

  push ldrmregs
  push ds
  push offset rmregs
  push sizeof rmregs
  call movefar
  add esp,16

  mov defsegs.appcs,cs                  ;load and extend defsegs structure
  mov defsegs.appds,ds
  mov defsegs.appss,ss
  mov word defsegs.rmcrtseg,0B800h
  mov word defsegs.rmvidseg,0A000h
  mov ax,es:[(lddefsegs+4)]
  mov defsegs.rmdosmemseg,ax
  mov ax,es:[(lddefsegs+6)]
  mov defsegs.crtseg,ax
  mov ax,es:[(lddefsegs+8)]
  mov defsegs.vidseg,ax
  mov ax,es:[(lddefsegs+10)]
  mov defsegs.dosmemseg,ax
  mov ax,es:[(lddefsegs+18)]
  mov defsegs.loadercs,ax
  mov defsegs.loaderds,es
  mov ax,es:[(lddefsegs+22)]
  mov defsegs.loaderss,ax
  mov ax,es:[(lddefsegs+24)]
  mov defsegs.pspseg,ax

  call allocdescriptor                  ;setup flat segment to provide access to address space 0..max-1 where max is systems memory at all
  jc @mend
  mov defsegs.flatmemseg,ax
  push ax
  push 0
  call setsegmentbase
  jnc @m1
  @m2:
  add esp,6
  jmp @mend
  @m1:
  pop eax
  mov eax,es:[(lddpmimeminfo+24)]       ;get memory available on system and set limit of flatseg
  shl eax,12
  dec eax
  push eax
  call setsegmentlimit
  jc @m2
  add esp,6

  call allocdescriptor                  ;setup BIOS data seg to provide access to BIOS data area
  jc @mend
  mov defsegs.biosmemseg,ax
  push ax
  push 400h
  call setsegmentbase
  jc @m2
  pop eax
  push 100h                             ;BIOS data area is 256 bytes
  call setsegmentlimit
  jc @m2
  add esp,6

  push word defsegs.pspseg              ;setup miscallineous RM segs
  call pmseg2rmseg
  pop cx
  jc @mend
  mov defsegs.rmpspseg,ax
  push word defsegs.loadercs
  call pmseg2rmseg
  pop cx
  jc @mend
  mov defsegs.rmloadercs,ax
  push word defsegs.loaderds
  call pmseg2rmseg
  pop cx
  jc @mend
  mov defsegs.rmloaderds,ax
  push word defsegs.loaderss
  call pmseg2rmseg
  pop cx
  jc @mend
  mov defsegs.rmloaderss,ax

  push offset dpmimeminfo               ;update memory info
  call getmeminfo
  add esp,4

  clc
  @mend:

endproc


;---OutDWord-------------------------------------------------------------------------------------------------------------------
;   return   : none

proc outdword(ptr offs, dword val)
  mov fs,defsegs.crtseg

  mov ebx,val
  mov ah,1Eh
  mov ecx,8

  mov edx,offs
  @mloop:
  mov al,bl
  and al,0Fh
  add al,30h
  cmp al,39h
  jna @mout1
  add al,7
  @mout1:
  shr ebx,4
  mov fs:[edx+ecx*2-2],ax
  dec ecx
  jnz @mloop
endproc


;---Halt-----------------------------------------------------------------------------------------------------------------------
;   return   : none


proc halt
  mov ax,4C00h
  int 21h
endproc


;---HaltError------------------------------------------------------------------------------------------------------------------
;   return   : none

proc halterror(word error)
  mov al,error
  mov ah,4Ch
  int 21h
endproc


;---FillChar-------------------------------------------------------------------------------------------------------------------
;   return   : none

proc fillchar(ptr buff, dword bytes, word val)
  mov bl,val                            
  mov bh,val
  shl ebx,16
  mov bl,val
  mov bh,val
  mov ecx,bytes
  mov eax,buff

  @mfill:

  test ecx,1
  jz @m1
  mov [eax+ecx-1],bl
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  mov [eax+ecx*2-2],bx
  @m2:
  shr ecx,1

  @m3:
  mov [eax+ecx*4-4],ebx
  dec ecx
  jnz @m3
endproc


;---FillWord-------------------------------------------------------------------------------------------------------------------
;   return   : none

proc fillword(ptr buff, dword bytes, word val)
  mov bx,val
  shl ebx,16
  mov bx,val
  mov ecx,bytes
  mov eax,buff
  jmp fillchar.mfill
endproc


;---FillDWord------------------------------------------------------------------------------------------------------------------
;   return   : none

proc filldword(ptr buff, dword bytes, dword val)
  mov ebx,val
  mov ecx,bytes
  mov eax,buff
  jmp fillchar.mfill
endproc


;---FillCharFar----------------------------------------------------------------------------------------------------------------
;   return   : none

proc fillcharfar(ptr48 buff, dword bytes, word val)
  mov bl,val
  mov bh,val
  shl ebx,16
  mov bl,val
  mov bh,val
  mov ecx,bytes
  mov eax,buff
  mov gs,[ebp+(offset buff+4)]

  @mfill:

  test ecx,1
  jz @m1
  mov gs:[eax+ecx-1],bl
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  mov gs:[eax+ecx*2-2],bx
  @m2:
  shr ecx,1

  @m3:
  mov gs:[eax+ecx*4-4],ebx
  dec ecx
  jnz @m3
endproc


;---FillWordFar----------------------------------------------------------------------------------------------------------------
;   return   : none

proc fillwordfar(ptr48 buff, dword bytes, word val)
  mov bx,val
  shl ebx,16
  mov bx,val
  mov ecx,bytes
  mov eax,buff
  mov gs,[ebp+(offset buff+4)]
  jmp fillcharfar.mfill
endproc


;---FillDWordFar---------------------------------------------------------------------------------------------------------------
;   return   : none

proc filldwordfar(ptr48 buff, dword bytes, dword val)
  mov ebx,val
  mov ecx,bytes
  mov eax,buff
  mov gs,[ebp+(offset buff+4)]
  jmp fillcharfar.mfill
endproc


;---AndBuffer------------------------------------------------------------------------------------------------------------------
;   return   : none

proc andbuffer(ptr48 buff, dword bytes, dword mask)
  mov ebx,mask
  mov ecx,bytes
  mov eax,buff
  mov gs,[ebp+(offset buff+4)]

  test ecx,1
  jz @m1
  and gs:[eax+ecx-1],bl
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  and gs:[eax+ecx*2-2],bx
  @m2:
  shr ecx,1

  @m3:
  and gs:[eax+ecx*4-4],ebx
  dec ecx
  jnz @m3
endproc


;---OrBuffer-------------------------------------------------------------------------------------------------------------------
;   return   : none

proc orbuffer(ptr48 buff, dword bytes, dword mask)
  mov ebx,mask
  mov ecx,bytes
  mov eax,buff
  mov gs,[ebp+(offset buff+4)]

  test ecx,1
  jz @m1
  or gs:[eax+ecx-1],bl
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  or gs:[eax+ecx*2-2],bx
  @m2:
  shr ecx,1

  @m3:
  or gs:[eax+ecx*4-4],ebx
  dec ecx
  jnz @m3
endproc


;---NandBuffer-----------------------------------------------------------------------------------------------------------------
;   return   : none

proc nandbuffer(ptr48 buff, dword bytes, dword mask)
  mov ebx,mask
  mov ecx,bytes
  mov eax,buff
  mov gs,[ebp+(offset buff+4)]

  test ecx,1
  jz @m1
  and gs:[eax+ecx-1],bl
  not byte gs:[eax+ecx-1]
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  and gs:[eax+ecx*2-2],bx
  not word gs:[eax+ecx*2-2]
  @m2:
  shr ecx,1

  @m3:
  and gs:[eax+ecx*4-4],ebx
  not dword gs:[eax+ecx*4-4]
  dec ecx
  jnz @m3
endproc


;---NorBuffer------------------------------------------------------------------------------------------------------------------
;   return   : none

proc norbuffer(ptr48 buff, dword bytes, dword mask)
  mov ebx,mask
  mov ecx,bytes
  mov eax,buff
  mov gs,[ebp+(offset buff+4)]

  test ecx,1
  jz @m1
  or gs:[eax+ecx-1],bl
  not byte gs:[eax+ecx-1]
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  or gs:[eax+ecx*2-2],bx
  not word gs:[eax+ecx*2-2]
  @m2:
  shr ecx,1

  @m3:
  or gs:[eax+ecx*4-4],ebx
  not dword gs:[eax+ecx*4-4]
  dec ecx
  jnz @m3
endproc


;---XorBuffer------------------------------------------------------------------------------------------------------------------
;   return   : none

proc xorbuffer(ptr48 buff, dword bytes, dword mask)
  mov ebx,mask
  mov ecx,bytes
  mov eax,buff
  mov gs,[ebp+(offset buff+4)]

  test ecx,1
  jz @m1
  xor gs:[eax+ecx-1],bl
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  xor gs:[eax+ecx*2-2],bx
  @m2:
  shr ecx,1

  @m3:
  xor gs:[eax+ecx*4-4],ebx
  dec ecx
  jnz @m3
endproc


;---XnorBuffer-----------------------------------------------------------------------------------------------------------------
;   return   : none

proc xnorbuffer(ptr48 buff, dword bytes, dword mask)
  mov ebx,mask
  mov ecx,bytes
  mov eax,buff
  mov gs,[ebp+(offset buff+4)]

  test ecx,1
  jz @m1
  xor gs:[eax+ecx-1],bl
  not byte gs:[eax+ecx-1]
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  xor gs:[eax+ecx*2-2],bx
  not word gs:[eax+ecx*2-2]
  @m2:
  shr ecx,1

  @m3:
  xor gs:[eax+ecx*4-4],ebx
  not dword gs:[eax+ecx*4-4]
  dec ecx
  jnz @m3
endproc


;---NotBuffer------------------------------------------------------------------------------------------------------------------
;   return   : none

proc notbuffer(ptr48 buff, dword bytes)
  mov ecx,bytes
  mov eax,buff
  mov gs,[ebp+(offset buff+4)]

  test ecx,1
  jz @m1
  not byte gs:[eax+ecx-1]
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  not word gs:[eax+ecx*2-2]
  @m2:
  shr ecx,1

  @m3:
  not dword gs:[eax+ecx*4-4]
  dec ecx
  jnz @m3
endproc


;---Move-----------------------------------------------------------------------------------------------------------------------
;   return   : none

proc move(ptr src, ptr dst, dword bytes)
  mov ebx,src
  mov eax,dst
  mov ecx,bytes

  test ecx,1
  jz @m1
  mov edx,[ebx+ecx-1]
  mov [eax+ecx-1],edx
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  mov edx,[ebx+ecx*2-2]
  mov [eax+ecx*2-2],edx
  @m2:
  shr ecx,1

  @m3:
  mov edx,[ebx+ecx*4-4]
  mov [eax+ecx*4-4],edx
  dec ecx
  jnz @m3
endproc


;---MoveFar--------------------------------------------------------------------------------------------------------------------
;   return   : none


proc movefar(ptr48 src, ptr48 dst, dword bytes)
  mov ebx,src
  mov fs,[ebp+(offset src+4)]
  mov eax,dst
  mov gs,[ebp+(offset dst+4)]
  mov ecx,bytes
  test ecx,1
  jz @m1
  mov edx,fs:[ebx+ecx-1]
  mov gs:[eax+ecx-1],edx
  @m1:
  shr ecx,1

  test ecx,1
  jz @m2
  mov edx,fs:[ebx+ecx*2-2]
  mov gs:[eax+ecx*2-2],edx
  @m2:
  shr ecx,1

  @m3:
  mov edx,fs:[ebx+ecx*4-4]
  mov gs:[eax+ecx*4-4],edx
  dec ecx
  jnz @m3
endproc


;---MoveReverse----------------------------------------------------------------------------------------------------------------
;   return   : none

proc movereverse(ptr src, ptr dst, dword bytes)
  mov ebx,src
  mov eax,dst
  xor ecx,ecx

  test bytes,1
  jz @m1
  mov edx,[ebx]
  mov [eax],edx
  add ebx,1
  add eax,1
  sub bytes,1
  @m1:

  test bytes,2
  jz @m2
  mov edx,[ebx]
  mov [eax],edx
  add ebx,2
  add eax,2
  sub bytes,2
  @m2:

  @m3:
  mov edx,[ebx+ecx*4]
  mov [eax+ecx*4],edx
  inc ecx
  sub bytes,4
  jnz @m3
endproc


;---MoveFarReverse-------------------------------------------------------------------------------------------------------------
;   return   : none

proc movefarreverse(ptr48 src, ptr48 dst, dword bytes)
  mov ebx,src
  mov fs,[ebp+(offset src+4)]
  mov eax,dst
  mov gs,[ebp+(offset dst+4)]
  xor ecx,ecx

  test bytes,1
  jz @m1
  mov edx,fs:[ebx]
  mov gs:[eax],edx
  add ebx,1
  add eax,1
  sub bytes,1
  @m1:

  test bytes,2
  jz @m2
  mov edx,fs:[ebx]
  mov gs:[eax],edx
  add ebx,2
  add eax,2
  sub bytes,2
  @m2:

  @m3:
  mov edx,fs:[ebx+ecx*4]
  mov gs:[eax+ecx*4],edx
  inc ecx
  sub bytes,4
  jnz @m3
endproc


;---GetMemInfo-----------------------------------------------------------------------------------------------------------------
;   return   : none

proc getmeminfo(ptr dst)
  push es
  push ds
  pop es
  mov edi,dst
  mov ax,500h
  int 31h
  pop es
endproc

;---AllocateDescriptor---------------------------------------------------------------------------------------------------------
;   return   : ax = selector of descriptor allocated if succeeds, CF set on error, ax = err_xxx

proc allocdescriptor
  xor ax,ax
  mov cx,1
  int 31h
endproc


;---AllocateDescriptors--------------------------------------------------------------------------------------------------------
;   return   : ax = selector of first descriptor allocated if succeeds, CF set on error, ax = err_xxx

proc allocdescriptors(word descriptors)
  xor ax,ax
  mov cx,descriptors
  int 31h
endproc


;---AllocateSpecDescriptor-----------------------------------------------------------------------------------------------------
;   return   : ax = selector of descriptor allocated if succeeds, CF set on error, ax = err_xxx

proc allocspecdescriptor(word descriptor)
  mov ax,0Dh
  mov bx,descriptor
  int 31h
endproc


;---FreeDescriptor-------------------------------------------------------------------------------------------------------------
;   return   : CF set on error, ax = err_xxx

proc freedescriptor(word descriptor)
  mov ax,1
  mov bx,descriptor
  int 31h
endproc


;---RMSeg2PMSeg----------------------------------------------------------------------------------------------------------------
;   return   : ax = selector of PM segment if succeeds, CF set on error, ax = err_xxx

proc rmseg2pmseg(word rmseg)
  mov ax,2
  mov bx,rmseg
  int 31h
endproc


;---PMSeg2RMSeg----------------------------------------------------------------------------------------------------------------
;   return   : ax = selector of RM segment if succeeds, CF set on error, ax = err_xxx

proc pmseg2rmseg(word pmseg)
  push pmseg
  call getsegmentbase
  pop cx
  jc @mend

  shl ax,4
  clc

  @mend:
endproc


;---GetSegmentBase-------------------------------------------------------------------------------------------------------------
;   return   : eax = segmentbase if succeeds, CF set on error, ax = err_xxx

proc getsegmentbase(word descriptor)
  mov ax,6
  mov bx,descriptor
  int 31h
  jc mend

  mov ax,cx
  shl eax,16
  mov ax,dx
  clc

  @mend:
endproc


;---SetSegmentBase-------------------------------------------------------------------------------------------------------------
;   return   : CF set on error, ax = err_xxx

proc setsegmentbase(word descriptor, dword base)
  mov ax,7
  mov bx,descriptor
  mov dx,base
  mov cx,[ebp+(offset base+2)]
  int 31h
endproc


;---GetSegmentLimit------------------------------------------------------------------------------------------------------------
;   return   : eax = segmentlimit if succeeds, CF set on error, ax = err_xxx

proc getsegmentlimit(word descriptor)
  mov bx,descriptor
  mov ax,0Bh

  sub esp,8
  mov edi,esp
  push ss
  pop es
  int 31h
  jnc @m1
  add esp,8
  stc
  pop ebp
  ret

  @m1:
  mov al,[ebp-2]
  and ax,0Fh
  shl eax,16
  test byte [ebp-2],80h
  jz @m2
  shl eax,12
  @m2:

  add esp,8
  clc
endproc


;---SetSegmentLimit------------------------------------------------------------------------------------------------------------
;   return   : CF set on error, ax = err_xxx

proc setsegmentlimit(word descriptor, dword limit)
  mov ax,8
  mov bx,descriptor
  mov ecx,limit
  cmp ecx,0FFFFFh
  jna @mnoalign                         ;make limit page alignet, since more than 1 MB
  or ecx,0FFFh
  @mnoalign:

  mov dx,cx
  shr ecx,16
  int 31h
endproc


;---RMInt----------------------------------------------------------------------------------------------------------------------
;   return   : CF set on error, ax = err_xxx

proc rmint(word int, dword rmregs)
  mov ax,300h
  mov bx,int                            ;note, that flags in bh should be zero, since DPMI 1.0 requires
  xor ecx,ecx
  push es
  push ds
  pop es
  mov edi,rmregs
  int 31h
  pop es
endproc
